/**
 *
 * @file
 *
 *  PLASMA is a software package provided by:
 *  University of Tennessee, US,
 *  University of Manchester, UK.
 *
 * @generated from /Volumes/ramdisk/build/plasma18.9.0/compute/zgeinv.c, normal z -> s, Thu Oct 18 17:45:35 2018
 *
 **/

#include "plasma.h"
#include "plasma_async.h"
#include "plasma_context.h"
#include "plasma_descriptor.h"
#include "plasma_internal.h"
#include "plasma_tuning.h"
#include "plasma_types.h"
#include "plasma_workspace.h"

/***************************************************************************//**
 *
 * @ingroup plasma_geinv
 *
 *  Performs the LU inversion of a matrix A.
 *
 *******************************************************************************
 *
 * @param[in] m
 *          The number of rows in the matrix A. m >= 0
 *
 * @param[in] n
 *          The number of columns in the matrix A. n >= 0.
 *
 * @param[in,out] pA
 *          On entry, the m-by-n matrix A to be inverted.
 *          On exit, the inverse of A.
 *
 * @param[in] lda
 *          The leading dimension of the array A. lda >= max(1,m).
 *
 * @param[out] ipiv
 *          The pivot indices; for 1 <= i <= min(m,n), row i of the
 *          matrix was interchanged with row ipiv(i).
 *
 *******************************************************************************
 *
 * @retval PlasmaSuccess successful exit
 * @retval  < 0 if -i, the i-th argument had an illegal value
 * @retval  > 0 if i, the leading minor of order i of A is not
 *          positive definite, so the factorization could not
 *          be completed, and the solution has not been computed.
 *
 *******************************************************************************
 *
 * @sa plasma_cgeinv
 * @sa plasma_dgeinv
 * @sa plasma_sgeinv
 *
 ******************************************************************************/
int plasma_sgeinv(int m, int n, float *pA, int lda, int *ipiv)
{
    // Get PLASMA context.
    plasma_context_t *plasma = plasma_context_self();
    if (plasma == NULL) {
        plasma_fatal_error("PLASMA not initialized");
        return PlasmaErrorNotInitialized;
    }

    // Check input arguments.
    if (m < 0) {
        plasma_error("illegal value of uplo");
        return -1;
    }
    if (n < 0) {
        plasma_error("illegal value of n");
        return -2;
    }
    if (lda < imax(1, n)) {
        plasma_error("illegal value of lda");
        return -4;
    }

    // quick return
    if (imax(n, 0) == 0 || imax(m, 0) == 0)
        return PlasmaSuccess;

    // Tune parameters.
    if (plasma->tuning)
        plasma_tune_geinv(plasma, PlasmaRealFloat, m, n);

    // Set tiling parameters.
    int nb = plasma->nb;

    // Initialize barrier.
    plasma_barrier_init(&plasma->barrier);

    // Create tile matrix.
    plasma_desc_t A;
    plasma_desc_t W;
    int retval;
    retval = plasma_desc_general_create(PlasmaRealFloat, nb, nb,
                                           m, n, 0, 0, m, n, &A);
    if (retval != PlasmaSuccess) {
        plasma_error("plasma_desc_general_create() failed");
        return retval;
    }

    retval = plasma_desc_general_create(PlasmaRealFloat, nb, nb,
                                        n, nb, 0, 0, n, nb, &W);
    if (retval != PlasmaSuccess) {
        plasma_error("plasma_desc_general_create() failed");
        plasma_desc_destroy(&A);
        return retval;
    }

    // Initialize sequence.
    plasma_sequence_t sequence;
    retval = plasma_sequence_init(&sequence);

    // Initialize request.
    plasma_request_t request;
    retval = plasma_request_init(&request);

    #pragma omp parallel
    #pragma omp master
    {
        // Translate to tile layout.
        plasma_omp_sge2desc(pA, lda, A, &sequence, &request);

        // Call the tile async function.
        plasma_omp_sgeinv(A, ipiv, W, &sequence, &request);

        // Translate back to LAPACK layout.
        plasma_omp_sdesc2ge(A, pA, lda, &sequence, &request);
    }

    // Free matrix A in tile layout.
    plasma_desc_destroy(&A);

    // Return status.
    int status = sequence.status;
    return status;
}

/***************************************************************************//**
 *
 * @ingroup plasma_geinv
 *
 *  Computes the inverse of a complex symmetric
 *  positive definite matrix A using the Cholesky factorization.
 *
 *******************************************************************************
 *
 * @param[in] A
 *          On entry, the symmetric positive definite matrix A.
 *          On exit, the upper or lower triangle of the (symmetric)
 *          inverse of A, overwriting the input factor U or L.
 *
 * @param[out] ipiv
 *          The pivot indices; for 1 <= i <= min(m,n), row i of the
 *          matrix was interchanged with row ipiv(i).
 *
 * @param[out] W
 *          Workspace of dimension (n, nb)
 *
 * @param[in] sequence
 *          Identifies the sequence of function calls that this call belongs to
 *          (for completion checks and exception handling purposes).  Check
 *          the sequence->status for errors.
 *
 * @param[out] request
 *          Identifies this function call (for exception handling purposes).
 *
 * @retval void
 *          Errors are returned by setting sequence->status and
 *          request->status to error values.  The sequence->status and
 *          request->status should never be set to PlasmaSuccess (the
 *          initial values) since another async call may be setting a
 *          failure value at the same time.
 *
 *******************************************************************************
 *
 * @sa plasma_sgeinv
 * @sa plasma_omp_sgeinv
 * @sa plasma_omp_cgeinv
 * @sa plasma_omp_dgeinv
 * @sa plasma_omp_sgeinv
 *
 ******************************************************************************/
void plasma_omp_sgeinv(plasma_desc_t A, int *ipiv, plasma_desc_t W,
                       plasma_sequence_t *sequence, plasma_request_t *request)
{
    // Get PLASMA context.
    plasma_context_t *plasma = plasma_context_self();
    if (plasma == NULL) {
        plasma_error("PLASMA not initialized");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }

    // Check input arguments.
    if (plasma_desc_check(A) != PlasmaSuccess) {
        plasma_error("invalid A");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }
    if (sequence == NULL) {
        plasma_error("NULL sequence");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }
    if (request == NULL) {
        plasma_error("NULL request");
        plasma_request_fail(sequence, request, PlasmaErrorIllegalValue);
        return;
    }

    // quick return
    if ((A.m == 0) || (A.n == 0)) {
        return;
    }

    // Factorize A.
    plasma_psgetrf(A, ipiv, sequence, request);

    // Invert triangular part.
    plasma_pstrtri(PlasmaUpper, PlasmaNonUnit, A, sequence, request);

    // Compute product of inverse of the upper and lower triangles.
    plasma_psgetri_aux(A, W, sequence, request);

    // Apply pivot.
    plasma_psgeswp(PlasmaColumnwise, A, ipiv, -1, sequence, request);
}
